import os
import json
import yaml

class YamlLoader(yaml.SafeLoader):    
    """
        Classe que implementa um loader customizado do YAML, que permite adicionar uma diretiva para incluir um YAML dentro de outro. Em tempo de execução os YAMLs vão para o mesmo objeto.

        Baseado em: https://gist.github.com/joshbode/569627ced3076931b02f#file-loader-py
    """
    def __init__(self, stream):
        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        super().__init__(stream)

def construct_include(loader, node):
    """
        Método executado internamente pela classe de loader do YAML. É ele que permite juntar os yml a partir da diretiva !include. Para mais detalhes ver a documentação do pyyaml.
    """

    filename = os.path.abspath(os.path.join(loader._root, loader.construct_scalar(node)))
    extension = os.path.splitext(filename)[1].lstrip('.')

    with open(filename, 'r') as f:
        if extension in ('yaml', 'yml'):
            return yaml.load(f, YamlLoader)
        elif extension in ('json', ):
            return json.load(f)
        else:
            return ''.join(f.readlines())

yaml.add_constructor('!include', construct_include, YamlLoader)


from pyspark.sql import SparkSession
from .utils import Utils 
from dnalib.log import log
from delta.tables import DeltaTable

class TableUtils:

    #  
    #   Description:
    #       This method formats a comment (for field or table).
    #
    #   Parameters:      
    #       comment_content = the content of the comment.
    # 
    @staticmethod
    def format_comment_content(comment_content):
        # remove whitespaces        
        if len(comment_content) > 0:
            value = comment_content.strip()
            # insert ponctuation at the end of string
            if value[-1] != '.':
                value += '.'
            # first digit in upper case        
            return value[0].upper() + value[1:]
        else: 
            return comment_content

    #  
    #   Description:
    #       This method formats a field to a hash type.
    #
    #   Parameters:      
    #       layer = field name to be a hashed field
    # 
    @staticmethod
    def hash_field_str(field):        
        return f"sha2({field}, 256) AS {field}"

    #  
    #   Description:
    #       This method loads a dataframe from path (usefull if table does not exists in catalog yet).
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    # 
    @staticmethod
    def load_df_from_lakehouse(layer, table_name):        
        root_path = Utils.lakehouse_path()
        try:
           df = Utils.spark_instance().read.load(f'{root_path}/{layer}/{table_name}')
        except:
            log(__name__).error(f"Unable to read '{root_path}/{layer}/{table_name}'")
            raise Exception(f"Unable to read '{root_path}/{layer}/{table_name}'")
        return df
    #  
    #   Description:
    #       This method gets schema table from path (usefull if table does not exists in catalog yet).
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    #     
    @staticmethod
    def table_schema_from_lakehouse(layer, table_name):                        
        return TableUtils.load_df_from_lakehouse(layer, table_name).schema
    
    #  
    #   Description:
    #       This method checks if a view exists from a table name in catalog.
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    # 
    @staticmethod
    def view_exists(layer, table_name):                                
        return Utils.spark_instance().catalog.tableExists(f"{layer}.{table_name}_vw")        
    
    #  
    #   Description:
    #       This method checks if a table exists in catalog.
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    # 
    @staticmethod
    def table_exists(layer, table_name):        
        #return Utils.spark_instance()._jsparkSession.catalog().tableExists(f"{layer}.{table_name}")
        try:
            DeltaTable.forName(Utils.spark_instance(), f"{layer}.{table_name}")
            return True
        except:
            return False
            
    #  
    #   Description:
    #       This method drop the table and remove data from lakehouse (delete path files).
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    #
    @staticmethod
    def drop_table(layer, table_name):
        Utils.spark_instance().sql(f"drop table if exists {layer}.{table_name}")
        is_removed = Utils.remove_table_from_lakehouse(layer, table_name)
        if not is_removed:
            log(__name__).warning(f"Table {layer}.{table_name} does not exists in lakehouse")  

    #  
    #   Description:
    #       This method drop the view from a table.
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    #
    @staticmethod
    def drop_view(layer, table_name):
        Utils.spark_instance().sql(f"drop view if exists {layer}.{table_name}_vw")        
                            


from pyspark.sql import SparkSession
from pyspark.dbutils import DBUtils
from dnalib.custom import YamlLoader
import yaml
from dnalib.log import log

class Utils:
    
    #  
    #   Description:
    #       This method removes a table from the lakehouse (it deletes the files).
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    #
    @staticmethod
    def remove_table_from_lakehouse(layer, table):
        table_path = Utils.lakehouse_path(layer, table)        
        return Utils.dbutils().fs.rm(table_path, True)        

    #  
    #   Description:
    #       This method returns spark instance.
    #
    @staticmethod
    def spark_instance():
        return SparkSession.builder.getOrCreate()

    #  
    #   Description:
    #       This method returns a new dbutils instance.
    #
    @staticmethod
    def dbutils():  
        return DBUtils(Utils.spark_instance())

    #  
    #   Description:
    #       This method returns storage var according to key vault secret scope.
    #
    @staticmethod
    def storage():
        SCOPE = "key-vault-secrets"
        return Utils.dbutils().secrets.get(scope=SCOPE, key="externalLocation")
    
    #  
    #   Description:
    #       This method returns the lakehouse path considering our storage configuration.
    #
    #   Parameters:      
    #       layer [optional] = table layer (bronze, silver, gold, diamond or export)
    #       table_name [optional] = the name of the table.
    #
    @staticmethod
    def lakehouse_path(layer=None, table_name=None):        
        storage = Utils.storage()
        if layer != None and table_name != None:
            return f"abfss://lakehouse@{storage}/{layer}/{table_name}"
        else:
            return f"abfss://lakehouse@{storage}"
        
    @staticmethod
    def landingzone_path(dir_path):
        storage = Utils.storage()
        return f"abfss://landingzone@{storage}/{dir_path}"
    
    @staticmethod
    def container_path(container_name, file_name):
        storage = Utils.storage()
        return f"abfss://{container_name}@{storage}/{file_name}"
        
    #  
    #   Description:
    #       This method loads and parse a yml file, it is based on custom YamlLoader (see source).
    #
    #   Parameters:      
    #       yml_file_path = the yaml file path.   
    #
    @staticmethod
    def safe_load_and_parse_yml(yml_file_path, write_workspace=False):
         # load yaml config file
        if write_workspace:
            yml_file_path = f'file:{yml_file_path}'
        try:
            yml_text = (
                Utils.spark_instance().read.text(
                    f'{yml_file_path}', wholetext=True)
                .first()[0]
            )
        except:
            log(__name__).error(f"Unable to read {yml_file_path}.")
            raise Exception(f"Unable to read {yml_file_path}.")

        ## parse the yaml
        try:
            parsed_yml = yaml.load(yml_text, YamlLoader)
        except:
            log(__name__).error(f"Unable to parse {yml_file_path}.")
            raise Exception(f"Unable to parse {yml_file_path}.")

        return parsed_yml
    
    @staticmethod
    def yaml_table_parameters(table_name, yml_file_path):
        table_config = None
        yaml_config = Utils.safe_load_and_parse_yml(yml_file_path)
        # Extração das chaves e regras de merge do YAML
        if table_name in yaml_config['tables']:
            table_config = yaml_config['tables'][table_name]            
        else:            
            log(__name__).error(f"Table {table_name} not found in yaml configuration.")
            raise Exception(f"Table {table_name} not found in yaml configuration.")
        return table_config
